#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

typedef struct gc_t {
  GC                 gcid;
  XGCValues          gcv;
  unsigned long int  gcm;
} gc_t;

typedef struct graf_data_t {
  gc_t          gc;      /* gc to use */
  char *        message; /* message to display */
  XFontStruct * font_struct;
} graf_data_t;

/* functions */
GC graf_makegc ();
void graf_resize ();
void graf_redraw ();
void graf_keydown ();
Widget create_graf_widget ();


/* default resources */
String defresources[] = { 
  "*graf.width:  300",
  "*graf.height: 200",
  NULL,
};

/* application-specific command line options */
XrmOptionDescRec options [] = {
  { "-message","*message", XrmoptionSepArg, NULL }};

main (argc, argv)
  int    argc;
  char  *argv[];
{
  Widget        toplevel;
  Widget        graf_widget;
  XtAppContext  app;

  /*
   * Initialize the intrinsics
   */
  toplevel = XtAppInitialize (&app, "Hello", 
			      options, XtNumber(options),
			      &argc, argv, 
			      defresources, NULL, 0);

  /*
   * create the workspace widget
   */

  graf_widget = create_graf_widget ( "graf", toplevel, NULL, 0);

  /* realize the widget and enter the main loop */
  XtRealizeWidget (toplevel);
  XtAppMainLoop(app);
}

static XtResource resources[] = {
  {XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
     XtOffset(graf_data_t *, font_struct), 
     XtRString, 
     "-*-charter-bold-r-*-*-*-240-*-*-p-*-iso8859-1" },
  {"message", "Message", XtRString, sizeof (char *),
     XtOffset(graf_data_t *, message), 
     XtRString, "Hello, World!" },
  {XtNforeground, XtCForeground, XtRPixel, sizeof (int),
     XtOffset(graf_data_t *, gc.gcv.foreground), 
     XtRString, "Black"}, 
  {XtNbackground, XtCBackground, XtRPixel, sizeof (int),
     XtOffset(graf_data_t *, gc.gcv.background),
     XtRString, "White"}
};

Widget
create_graf_widget (name, parent, args, num_args)
    String     name;
    Widget     parent;
    ArgList    args;
    Cardinal   num_args;
{
  Widget        graf_widget;
  graf_data_t * graf_data;

  graf_widget = XtCreateManagedWidget 
                   (name,
		    widgetClass,
		    parent, args, num_args );
  
  graf_data = (graf_data_t * ) calloc (1, sizeof (graf_data_t));

  XtGetApplicationResources (parent, graf_data, resources,
                 XtNumber (resources), NULL, 0);

  /* expose event handler */
  XtAddEventHandler ( graf_widget, ExposureMask, 
       False, graf_redraw, graf_data);
  /* key and button press handler */
  XtAddEventHandler ( graf_widget, ButtonPressMask|KeyPressMask,
       False, graf_keydown, graf_data);
  return graf_widget;

}

/* ARGSUSED */
void 
graf_redraw (widget, data, ev )
  Widget widget;
  graf_data_t * data;
  XEvent * ev;
{
  Display * dpy = XtDisplay (widget);
  Window w      = XtWindow  (widget);
  GC gc         = graf_makegc ( widget, data );
  
  Position x,y;
  Dimension width, height;
  Dimension wwidth, wheight;
  static Arg arg[2];
  XCharStruct size;
  int direction, ascent, descent;
  Dimension border = 10;
  Dimension margin = 20;
  
  /* look up widget size */
  XtSetArg (arg[0], XtNwidth,  &wwidth);
  XtSetArg (arg[1], XtNheight, &wheight);
  XtGetValues (widget, arg, 2);

  /* look up string size of message */
  XTextExtents (data->font_struct, 
		data->message, strlen (data->message),
		&direction, &ascent, &descent, &size);
  
  /* center the message in the window */
  width = size.rbearing - size.lbearing;
  height = ascent + descent;
  x = ((wwidth  - width) / 2) - size.lbearing;
  y = ((wheight - height) / 2) + ascent;
  
  /* draw the message */
  XDrawString (dpy, w, gc, x, y, 
	       data->message, strlen (data->message));
  
  /* position a wide border */
  x += size.lbearing;
  x -= (margin + border/2);
  width += border + (2 * margin);

  y -= ascent;
  y -= (margin + border/2);
  height += border + (2 * margin);
  
  /* draw the border */
  XSetLineAttributes (dpy, gc, border, LineSolid, 
		      CapRound, JoinRound);
  XDrawRectangle (dpy, w, gc, x,y, width, height);

}

GC 
graf_makegc ( widget, data )
  Widget widget;
  graf_data_t * data;
{
  XSetWindowAttributes xswa;
  if (data->gc.gcid == NULL) {

    /* initialize */
    Display * dpy = XtDisplay (widget);
    Window win    = XtWindow  (widget);

    /* set window gravity for centered graphics */
    xswa.bit_gravity = ForgetGravity;
    XChangeWindowAttributes(dpy, win, CWBitGravity, xswa);

    /* make and initialize the drawing GC */
    data->gc.gcid = XCreateGC (dpy, win, 0, 0);
    if (data->font_struct->fid) 
       XSetFont (dpy, data->gc.gcid, data->font_struct->fid);
    XSetForeground (dpy, data->gc.gcid,
		    data->gc.gcv.foreground);
    XSetBackground (dpy, data->gc.gcid, 
		    data->gc.gcv.background);

  }
  return data->gc.gcid;
}


/* ARGSUSED */
void graf_keydown ( widget, data, ev )
  Widget widget;
  graf_data_t * data;
  XEvent * ev;
{
  exit (0);
}

